#include "pgn2pov.h"

struct cell {
	unsigned piece	:3;
	unsigned color	:1;
} board[8][8];
unsigned short WhiteCaptured[15], BlackCaptured[15];
short WhiteCapIdx, BlackCapIdx;
FILE *pfIn, *pfOut;
int game;
double dMove;

int main(int argc, char **argv)
{
	char input[MAX_PATH], output[MAX_PATH];
	int ret, iArgCtr;
	
	game=1;
	dMove=1000;
	iArgCtr=2;

	if (argc<2)
	{
		printf("\nUsage:\n\tpgn2pov inputFile [outputFile] [-gn] [-mn]");
		printf("\n\t\t-gn\tconvert game #n");
		printf("\n\t\t-mn\tstop after parsing move #n");
		return 0x10;
	}
	strcpy(input,argv[1]);
	if ((argc<3) || (argv[2][0]=='-'))
	{
		strcpy(output, input);
		strcat(output,".inc");
	}
	else if (argc>2)
	{
		strcpy(output,argv[2]);
		iArgCtr=3;
	}
	for (;iArgCtr<argc;iArgCtr++)
	{
		if (argv[iArgCtr][0]!='-')
		{
			printf("\nInvalid flag %s\nAborting...",argv[iArgCtr]);
			return 0x11;
		}
		switch (argv[iArgCtr][1])
		{
			case 'g':
			case 'G':
				game=atoi(&argv[iArgCtr][2]);
				if (game<1)
				{
					printf("\nInvalid game number %i",game);
					return 0x12;
				}
				break;
			case 'm':
			case 'M':
				dMove=(atof(&argv[iArgCtr][2])+1)/2.0+0.5;
				break;
			default:
				printf("\nIgnoring unknown flag %s",argv[iArgCtr]);
				break;
		}
	}

	printf("\nConverting %s to %s, game #%i, through move %.1f",input, output, game, dMove-1);

	if ((pfIn=fopen(input,"r"))==NULL)
	{
		printf("\nInput file %s could not be opened for read",input);
		return -2;
	}
	if ((pfOut=fopen(output,"w"))==NULL)
	{
		printf("\nOutput file %s could not be opened for write",output);
		fclose(pfIn);
		return -3;
	}

	if ((ret=initBoard())==0)
	{
		if ((ret=readBoard())==0)
		{
			ret=outputBoard();
		}
	}

	fclose(pfIn);
	fclose(pfOut);

	return ret;
}

int initBoard(void)
{
	int i, ii;

	for (i=0;i<8;i++)
		board[i][6].piece=board[i][1].piece=PAWN;
	board[A][0].piece=board[A][7].piece=board[H][0].piece=board[H][7].piece=ROOK;
	board[B][0].piece=board[B][7].piece=board[G][0].piece=board[G][7].piece=KNIGHT;
	board[C][0].piece=board[C][7].piece=board[F][0].piece=board[F][7].piece=BISHOP;
	board[D][0].piece=board[D][7].piece=QUEEN;
	board[E][0].piece=board[E][7].piece=KING;
	for (i=0;i<8;i++)
	{
		for (ii=0;ii<2;ii++)
		{
			board[i][ii].color=WHITE;
			board[i][7-ii].color=BLACK;
		}
	}
	return 0;
}

int outputBoard(void)
{
	int i, ii;
	cell me;

	for (i=0;i<8;i++)
	{
		for (ii=0;ii<8;ii++)
		{
			me=board[i][ii];
			switch (me.piece)
			{
				case PAWN:
					fprintf(pfOut,"\n_Pawn(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				case ROOK:
					fprintf(pfOut,"\n_Rook(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				case KNIGHT:
					fprintf(pfOut,"\n_Knight(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				case BISHOP:
					fprintf(pfOut,"\n_Bishop(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				case QUEEN:
					fprintf(pfOut,"\n_Queen(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				case KING:
					fprintf(pfOut,"\n_King(\"%c%i\",%i)",i+97,ii+1,me.color);
					break;
				default:
					break;
			}
		}
	}
	for (i=0;i<WhiteCapIdx;i++)
	{
		switch (WhiteCaptured[i])
		{
			case PAWN:
				fprintf(pfOut,"\n_Pawn(\"i%i\",%i)",i+1,WHITE);
				break;
			case ROOK:
				fprintf(pfOut,"\n_Rook(\"i%i\",%i)",i+1,WHITE);
				break;
			case KNIGHT:
				fprintf(pfOut,"\n_Knight(\"i%i\",%i)",i+1,WHITE);
				break;
			case BISHOP:
				fprintf(pfOut,"\n_Bishop(\"i%i\",%i)",i+1,WHITE);
				break;
			case QUEEN:
				fprintf(pfOut,"\n_Queen(\"i%i\",%i)",i+1,WHITE);
				break;
			case KING:
				fprintf(pfOut,"\n_King(\"i%i\",%i)",i+1,WHITE);
				break;
			default:
				break;
		}
	}
	for (i=0;i<BlackCapIdx;i++)
	{
		switch (BlackCaptured[i])
		{
			case PAWN:
				fprintf(pfOut,"\n_Pawn(\"`%i\",%i)",8-i,BLACK);
				break;
			case ROOK:
				fprintf(pfOut,"\n_Rook(\"`%i\",%i)",8-i,BLACK);
				break;
			case KNIGHT:
				fprintf(pfOut,"\n_Knight(\"`%i\",%i)",8-i,BLACK);
				break;
			case BISHOP:
				fprintf(pfOut,"\n_Bishop(\"`%i\",%i)",8-i,BLACK);
				break;
			case QUEEN:
				fprintf(pfOut,"\n_Queen(\"`%i\",%i)",8-i,BLACK);
				break;
			case KING:
				fprintf(pfOut,"\n_King(\"`%i\",%i)",8-i,BLACK);
				break;
			default:
				break;
		}
	}
	return 0;
}

int readBoard(void)
{
	char line[500], *move, *ptr, *white, *black;
	char delim[]=". \n\r";
	int moveCtr=1, iGameCtr;
	bool insideVar=false;

	line[0]='[';

	for (iGameCtr=1;iGameCtr<=game;iGameCtr++)
	{
		while (((line[0]=='[') || (line[0]==10)) && (!feof(pfIn)))
			fgets(line,500,pfIn);
		if (ferror(pfIn))
		{
			printf("\nGame #%i does not exist in this file",game);
			return 0x100;
		}
		if (iGameCtr<game)
		{
			while ((line[0]!='[') && (!feof(pfIn)))
				fgets(line,500,pfIn);
			if (feof(pfIn))
			{
				printf("\nGame #%i does not exist in this file",game);
				return 0x100;
			}
		}
	}

	while (pfIn)
	{
		ptr=line;
		while (ptr)
		{
			//get next move
			move=strtok(ptr,delim);
			if (move==NULL)
				break;
			ptr=move+strlen(move)+1;
			if (move[0]=='{')
			{
				skipVar(&ptr);
				move=strtok(ptr,delim);
			}
			if (move[0]=='*')
				return 0;
			else if (strstr(move,"-"))
				return 0;
			else if (atoi(move)!=moveCtr)
			{
				printf("\nMove %s found, expecting %i",move,moveCtr);
				//according to strict pgn file format specs, this doesn't matter
				//return 0x101;
			}
ReGetWhite:
			white=strtok(ptr,delim);
			if (white==NULL)
			{
				if (!feof(pfIn))
				{
					fgets(line,500,pfIn);
					ptr=line;
				}
				else
					return 0x102;
				goto ReGetWhite;
			}
			ptr=white+strlen(white)+1;
			if (white[0]=='{')
			{
				skipVar(&ptr);
				move=strtok(ptr,delim);
			}
			if (moveCtr>=dMove)
				return 0;
			if (movePiece(white,WHITE))
				return 0;
ReGetBlack:
			black=strtok(ptr,delim);
			if (black==NULL)
			{
				if (!feof(pfIn))
				{
					fgets(line,500,pfIn);
					ptr=line;
				}
				else
					return 0x102;
				goto ReGetBlack;
			}
			ptr=black+strlen(black)+1;
			if (black[0]=='{')
			{
				skipVar(&ptr);
				move=strtok(ptr,delim);
			}
			if (moveCtr+0.5>=dMove)
				return 0;
			if (movePiece(black,BLACK))
				return 0;
			moveCtr++;
		}
		if (!feof(pfIn))
			fgets(line,500,pfIn);
		else
			break;
	}
	printf("\nUnexpected end-of-file found");
	return 0x102;
}

int movePiece(char *move, int color)
{
	int type, rank, file, moveDir;
	short i, ii;
	int file1, rank1;
	bool capture=false;
	short fromRank=-1, fromFile=-1;
	int xDir, yDir;

	switch (move[0])
	{
		case '*':
			return -1;
		case 'N':
			type=KNIGHT;
			move++;
			break;
		case 'R':
			type=ROOK;
			move++;
			break;
		case 'B':
			type=BISHOP;
			move++;
			break;
		case 'Q':
			type=QUEEN;
			move++;
			break;
		case 'K':
			type=KING;
			move++;
			break;
		case 'o':
		case 'O':
			//castle
			if (strlen(move)>3)
			{
				//queen-side castle
				board[2][(color==WHITE) ? 0 : 7].piece=KING;
				board[3][(color==WHITE) ? 0 : 7].piece=ROOK;
				board[0][(color==WHITE) ? 0 : 7].piece=EMPTY;
				board[4][(color==WHITE) ? 0 : 7].piece=EMPTY;
				board[2][(color==WHITE) ? 0 : 7].color=color;
				board[3][(color==WHITE) ? 0 : 7].color=color;
			}
			else
			{
				//king-side castle
				board[6][(color==WHITE) ? 0 : 7].piece=KING;
				board[5][(color==WHITE) ? 0 : 7].piece=ROOK;
				board[4][(color==WHITE) ? 0 : 7].piece=EMPTY;
				board[7][(color==WHITE) ? 0 : 7].piece=EMPTY;
				board[6][(color==WHITE) ? 0 : 7].color=color;
				board[5][(color==WHITE) ? 0 : 7].color=color;
			}
			return 0;
			break;
		case 'P':
		default:
			type=PAWN;
			break;
	}
	if (move[0]=='x')
	{
		capture=true;
		move++;
	}
	else if (move[1]=='x')
	{
		capture=true;
		if (isdigit(move[0]))
			fromRank=move[0]-1-48;
		else
			fromFile=move[0]-97;
		move+=2;
	}
	else if (move[2]=='x')
	{
		capture=true;
		fromFile=move[0]-97;
		fromRank=move[1]-1-48;
		move+=3;
	}
	file=move[0]-97;
	if (!isdigit(move[1]))
	{
		fromFile=file;
		move++;
		if (isdigit(move[0]))
		{
			fromRank=move[0]-1-48;
			move++;
		}
		file=move[0]-97;
	}
	rank=move[1]-1-48;
	if (capture)
	{
		if (color==WHITE)
			BlackCaptured[BlackCapIdx++]=board[file][rank].piece;
		else
			WhiteCaptured[WhiteCapIdx++]=board[file][rank].piece;
	}
	move+=2;

	//where did this piece come from?
	bool foundPiece=false;

	if ((fromFile>-1) && (fromRank>-1))
		board[fromFile][fromRank].piece=EMPTY;
	else
	{
		switch (type)
		{
			case KNIGHT:
				//one quadrant at a time
				for (i=0;i<4;i++)
				{
					sign(i*1.5707963268+0.785398163398,&xDir,&yDir);
					file1=file+xDir;
					rank1=rank+yDir*2;
					if (fromFile>-1)
						file1=fromFile;
					else if (fromRank>-1)
						rank1=fromRank;
					if ((file1<8) && (rank1<8) && (file1>=0) && (rank1>=0))
					{
						if ((board[file1][rank1].color==(unsigned) color) &&
							(board[file1][rank1].piece==KNIGHT))
						{
							board[file1][rank1].piece=EMPTY;
							break;
						}
					}
					file1=file+xDir*2;
					rank1=rank+yDir;
					if (fromFile>-1)
						file1=fromFile;
					else if (fromRank>-1)
						rank1=fromRank;
					if ((file1<8) && (rank1<8) && (file1>=0) && (rank1>=0))
					{
						if ((board[file1][rank1].color==(unsigned) color) &&
							(board[file1][rank1].piece==KNIGHT))
						{
							board[file1][rank1].piece=EMPTY;
							break;
						}
					}
				}
				break;
			case PAWN:
				moveDir=color*2-1;
				if (capture)
				{
					if (fromFile>-1)
						board[fromFile][rank+moveDir].piece=EMPTY;
					//check for en-passant capture
					if (board[file][rank].piece==EMPTY)
					{
						board[file][rank+moveDir].piece=EMPTY;
						if (color==WHITE)
							BlackCaptured[BlackCapIdx++]=PAWN;
						else
							WhiteCaptured[WhiteCapIdx++]=PAWN;
					}
					break;
				}
				if (board[file][rank+moveDir].piece==PAWN)
					board[file][rank+moveDir].piece=EMPTY;
				else
					board[file][rank+moveDir*2].piece=EMPTY;
				break;
			case KING:
				for (i=0;i<8;i++)
				{
					sign(i*0.785398163398,&xDir,&yDir);
					file1=file+xDir;
					rank1=rank+yDir;
					if ((file1<8) && (rank1<8) && (file1>=0) && (rank1>=0))
					{
						if ((board[file1][rank1].color==(unsigned) color) &&
							(board[file1][rank1].piece==KING))
						{
							board[file1][rank1].piece=EMPTY;
							break;
						}
					}
				}
				break;
			case QUEEN:
				//this would have been easy except for pawn promotion
				for (i=0;i<8;i++)
				{
					sign(i*0.785398163398,&xDir,&yDir);
					rank1=0; file1=0;
					for (ii=1;(rank1<8) && (rank1>-1) && (file1<8) && (file1>-1);ii++)					
					{
						file1=file+xDir*ii;
						rank1=rank+yDir*ii;
						if (!((file1<8) && (rank1<8) && (file1>=0) && (rank1>=0)))
							break;
						if (fromFile>-1)
							file1=fromFile;
						else if (fromRank>-1)
							rank1=fromRank;
						if ((board[file1][rank1].piece==QUEEN) && 
							(board[file1][rank1].color==(unsigned) color))
						{
							board[file1][rank1].piece=EMPTY;
							foundPiece=true;
							break;
						}
						else if (board[file1][rank1].piece)
							break;
					}
					if (foundPiece)
						break;
				}
				break;
			case BISHOP:
				//check one direction at a time
				for (i=0;i<4;i++)
				{
					sign(i*1.5707963268+0.785398163398,&xDir,&yDir);
					rank1=0; file1=0;
					for (ii=1;(rank1<8) && (rank1>-1) && (file1<8) && (file1>-1);ii++)					
					{
						file1=file+xDir*ii;
						rank1=rank+yDir*ii;
						if (!((file1<8) && (rank1<8) && (file1>=0) && (rank1>=0)))
							break;
						if (fromFile>-1)
							file1=fromFile;
						else if (fromRank>-1)
							rank1=fromRank;
						if ((board[file1][rank1].piece==BISHOP) &&
							(board[file1][rank1].color==(unsigned) color))
						{
							board[file1][rank1].piece=EMPTY;
							foundPiece=true;
							break;
						}
						else if (board[file1][rank1].piece)
							break;
					}
					if (foundPiece)
						break;
				}
				break;
			case ROOK:
				//check one direction at a time
				for (i=0;i<4;i++)
				{
					sign(i*1.5707963268,&xDir,&yDir);
					rank1=0; file1=0;
					for (ii=1;(rank1<8) && (rank1>-1) && (file1<8) && (file1>-1);ii++)
					{
						file1=file+xDir*ii;
						rank1=rank+yDir*ii;
						if (!((rank1<8) && (rank1>-1) && (file1<8) && (file1>-1)))
							break;
						if (fromFile>-1)
							file1=fromFile;
						else if (fromRank>-1)
							rank1=fromRank;
						if ((board[file1][rank1].piece==ROOK) && (board[file1][rank1].color==(unsigned) color))
						{
							board[file1][rank1].piece=EMPTY;
							foundPiece=true;
							break;
						}
						else if (board[file1][rank1].piece)
							break;
					}
					if (foundPiece)
						break;
				}
				break;
		}
	}
	board[file][rank].color=color;
	if (move[0]=='=')
		move++;
	switch (move[0])//pawn promotion
	{
		case 'N':
			type=KNIGHT;
			break;
		case 'R':
			type=ROOK;
			break;
		case 'B':
			type=BISHOP;
			break;
		case 'Q':
			type=QUEEN;
			break;
		default:
			break;
	}
	board[file][rank].piece=type;

	return 0;
}

int skipVar(char **line)
{
	*line=strtok(*line,"}");
	while (*line==NULL)
	{
		if (!feof(pfIn))
			fgets(*line,500,pfIn);
		else
			return 0x1000;
		*line=strtok(*line,"}");
	}
	*line=(*line)+strlen(*line)+1;

	return 0;
}

inline void sign(double angle, int *x, int *y)
{
	double c=cos(angle);
	double s=sin(angle);

	if (fabs(c)<1.0e-10)
		*x=0;
	else
		*x=(int) (c/fabs(c));

	if (fabs(s)<1.0e-10)
		*y=0;
	else
		*y=(int) (s/fabs(s));
}